@@ -1,31 +1,22 @@
 /* agent.c - Talking to gpg-agent.
-   Copyright (C) 2006, 2007, 2008, 2015 g10 Code GmbH
-
-   This file is part of Scute.
-
-   Scute is free software; you can redistribute it and/or modify it
-   under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
-
-   Scute is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with Scute; if not, write to the Free Software Foundation,
-   Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
-
-   In addition, as a special exception, g10 Code GmbH gives permission
-   to link this library: with the Mozilla Foundation's code for
-   Mozilla (or with modified versions of it that use the same license
-   as the "Mozilla" code), and distribute the linked executables.  You
-   must obey the GNU General Public License in all respects for all of
-   the code used other than "Mozilla".  If you modify this file, you
-   may extend this exception to your version of the file, but you are
-   not obligated to do so.  If you do not wish to do so, delete this
-   exception statement from your version.  */
+ * Copyright (C) 2006, 2007, 2008, 2015, 2019 g10 Code GmbH
+ *
+ * This file is part of Scute.
+ *
+ * Scute is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation; either version 2.1 of
+ * the License, or (at your option) any later version.
+ *
+ * Scute is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this program; if not, see <https://gnu.org/licenses/>.
+ * SPDX-License-Identifier: LGPL-2.1-or-later
+ */
 
 #if HAVE_CONFIG_H
 #include <config.h>
@@ -39,11 +30,12 @@
 #include <stdarg.h>
 
 #ifdef HAVE_W32_SYSTEM
-#define PATHSEP_C ';'
-#define WINVER 0x0500  /* Required for AllowSetForegroundWindow.  */
-#include <windows.h>
+# define PATHSEP_C ';'
+# define WINVER 0x0500  /* Required for AllowSetForegroundWindow.  */
+# include <winsock2.h>
+# include <windows.h>
 #else
-#define PATHSEP_C ':'
+# define PATHSEP_C ':'
 #endif
 
 #include <assuan.h>
@@ -57,11 +49,7 @@
 
 
 /* The global agent context.  */
-static assuan_context_t agent_ctx = NULL;
-
-/* The version number of the agent.  */
-static int agent_version_major;
-static int agent_version_minor;
+static assuan_context_t agent_ctx;
 
 
 
@@ -73,8 +61,8 @@ gnupg_allow_set_foregound_window (pid_t pid)
     return;
 #ifdef HAVE_W32_SYSTEM
   else if (!AllowSetForegroundWindow (pid))
-    DEBUG (DBG_CRIT, "AllowSetForegroundWindow(%lu) failed: %i\n",
-	   (unsigned long)pid, GetLastError ());
+    DEBUG (DBG_CRIT, "AllowSetForegroundWindow(%lu) failed: %u\n",
+	   (unsigned long)pid, (unsigned int)GetLastError ());
 #endif
 }
 
@@ -86,32 +74,26 @@ agent_connect (assuan_context_t *ctx_r)
 {
   gpg_error_t err = 0;
   assuan_context_t ctx = NULL;
-  char buffer[255];
-  FILE *p;
+  char buffer[512];
 
-  /* Use gpg-connect-agent to obtain the socket name
-   * directly from the agent itself. */
-  snprintf (buffer, sizeof buffer, "%s 'GETINFO socket_name' /bye",
-            get_gpg_connect_agent_path ());
-#ifdef HAVE_W32_SYSTEM
-  p = _popen (buffer, "r");
-#else
-  p = popen (buffer, "r");
-#endif
-  if (p)
+  /* Use gpgconf to make sure that gpg-agent is started and to obtain
+   * the socket name.  For older version of gnupg we will fallback to
+   * using two gpgconf commands with the same effect.  */
+  snprintf (buffer, sizeof buffer, "%s --show-socket --launch gpg-agent",
+            get_gpgconf_path ());
+  err = read_first_line (buffer, buffer, sizeof buffer);
+  if (gpg_err_code (err) == GPG_ERR_NO_AGENT && is_gnupg_older_than (2, 2, 14))
     {
-      int ret;
-
-      ret = fscanf (p, "D %254s\nOK\n", buffer);
-      if (ret == EOF)       /* I/O error? */
-        err = gpg_error_from_errno (errno);
-      else if (ret != 1)    /* Unexpected reply */
-        err = gpg_error (GPG_ERR_NO_AGENT);
-
-      pclose (p);
+      snprintf (buffer, sizeof buffer, "%s --launch gpg-agent",
+                get_gpgconf_path ());
+      err = read_first_line (buffer, NULL, 0);
+      if (!err)
+        {
+          snprintf (buffer, sizeof buffer, "%s --list-dirs agent-socket",
+                    get_gpgconf_path ());
+          err = read_first_line (buffer, buffer, sizeof buffer);
+        }
     }
-  else
-    err = gpg_error_from_errno (errno);
 
   /* Then connect to the socket we got. */
   if (!err)
@@ -186,28 +168,6 @@ agent_simple_cmd (assuan_context_t ctx, const char *fmt, ...)
 }
 
 
-/* Read and stroe the agent's version number.  */
-static gpg_error_t
-read_version_cb (void *opaque, const void *buffer, size_t length)
-{
-  char version[20];
-  const char *s;
-
-  (void) opaque;
-
-  if (length > sizeof (version) -1)
-    length = sizeof (version) - 1;
-  strncpy (version, buffer, length);
-  version[length] = 0;
-
-  agent_version_major = atoi (version);
-  s = strchr (version, '.');
-  agent_version_minor = s? atoi (s+1) : 0;
-
-  return 0;
-}
-
-
 /* Configure the GPG agent at connection CTX.  */
 static gpg_error_t
 agent_configure (assuan_context_t ctx)
@@ -307,21 +267,51 @@ agent_configure (assuan_context_t ctx)
   if (err && gpg_err_code (err) != GPG_ERR_UNKNOWN_OPTION)
     return err;
 
-  err = assuan_transact (ctx, "GETINFO version",
-                         read_version_cb, NULL,
-                         NULL, NULL, NULL, NULL);
-  if (gpg_err_code (err) == GPG_ERR_UNKNOWN_OPTION)
-    err = 0;
-  else if (err)
-    return err;
-
-
   return err;
 }
 
 
+/* Check for a broken pipe, that is a lost connection to the agent.
+ * Update the gloabls so that a re-connect is done the next time.
+ * Returns ERR or the modified code GPG_ERR_NO_AGENT.  */
+static gpg_error_t
+check_broken_pipe (gpg_error_t err)
+{
+  /* Note that Scute _currently_ uses GPG_ERR_SOURCE_ANY.  */
+  if (gpg_err_code (err) == GPG_ERR_EPIPE
+      && gpg_err_source (err) == GPG_ERR_SOURCE_ANY)
+    {
+      DEBUG (DBG_INFO, "Broken connection to the gpg-agent");
+      scute_agent_finalize ();
+      err = gpg_error (GPG_ERR_NO_AGENT);
+    }
+  return err;
+}
+
+
+/* If the connection to the agent was lost earlier and detected by
+ * check_broken_pipe we try to reconnect.  */
+static gpg_error_t
+ensure_agent_connection (void)
+{
+  gpg_error_t err;
+
+  if (agent_ctx)
+    return 0;  /* Connection still known.  */
+
+  DEBUG (DBG_INFO, "Re-connecting to gpg-agent");
+  err = agent_connect (&agent_ctx);
+  if (err)
+    return err;
+
+  err = agent_configure (agent_ctx);
+  return check_broken_pipe (err);
+}
+
+
 /* Try to connect to the agent via socket.  Handle the server's
-   initial greeting.  */
+   initial greeting.  This is used only once when SCute is loaded.
+   Re-connection is done using ensure_agent_connection.  */
 gpg_error_t
 scute_agent_initialize (void)
 {
@@ -346,14 +336,6 @@ scute_agent_initialize (void)
 }
 
 
-int
-scute_agent_get_agent_version (int *minor)
-{
-  *minor = agent_version_minor;
-  return agent_version_major;
-}
-
-
 
 /* Return a new malloced string by unescaping the string S.  Escaping
    is percent escaping and '+'/space mapping.  A binary nul will
@@ -416,6 +398,21 @@ unhexify_fpr (const char *hexstr, unsigned char *fpr)
   return 1;
 }
 
+/* Return true if HEXSTR is a valid keygrip.  */
+static unsigned int
+hexgrip_valid_p (const char *hexstr)
+{
+  const char *s;
+  int n;
+
+  for (s=hexstr, n=0; hexdigitp (s); s++, n++)
+    ;
+  if ((*s && *s != ' ') || n != 40)
+    return 0; /* Bad keygrip */
+  else
+    return 1; /* Valid.  */
+}
+
 
 /* Take the serial number from LINE and return it verbatim in a newly
    allocated string.  We make sure that only hex characters are
@@ -448,22 +445,71 @@ scute_agent_release_card_info (struct agent_card_info_s *info)
     return;
 
   free (info->serialno);
+  free (info->dispserialno);
+  free (info->cardtype);
   free (info->disp_name);
   free (info->disp_lang);
   free (info->pubkey_url);
   free (info->login_data);
+  while (info->kinfo)
+    {
+      key_info_t ki = info->kinfo->next;
+      free (info->kinfo);
+      info->kinfo = ki;
+    }
 
   memset (info, 0, sizeof (*info));
 }
 
 
+/* Return the key info object for the key KEYREF.  If it is not found
+ * NULL is returned.  */
+key_info_t
+scute_find_kinfo (agent_card_info_t info, const char *keyref)
+{
+  key_info_t kinfo;
+
+  for (kinfo = info->kinfo; kinfo; kinfo = kinfo->next)
+    if (!strcmp (kinfo->keyref, keyref))
+      return kinfo;
+  return NULL;
+}
+
+
+/* Create a new key info object with KEYREF.  All fields but the
+ * keyref are zeroed out.  The created object is appended to the list
+ * at INFO. */
+static key_info_t
+create_kinfo (agent_card_info_t info, const char *keyref)
+{
+  key_info_t kinfo, ki;
+
+  kinfo = calloc (1, sizeof *kinfo + strlen (keyref));
+  if (!kinfo)
+    return NULL;
+  strcpy (kinfo->keyref, keyref);
+
+  if (!info->kinfo)
+    info->kinfo = kinfo;
+  else
+    {
+      for (ki=info->kinfo; ki->next; ki = ki->next)
+        ;
+      ki->next = kinfo;
+    }
+  return kinfo;
+}
+
+
 /* FIXME: We are not returning out of memory errors.  */
 static gpg_error_t
 learn_status_cb (void *opaque, const char *line)
 {
-  struct agent_card_info_s *parm = opaque;
+  agent_card_info_t parm = opaque;
   const char *keyword = line;
   int keywordlen;
+  key_info_t kinfo;
+  const char *keyref;
   int i;
 
   for (keywordlen = 0; *line && !spacep (line); line++, keywordlen++)
@@ -473,10 +519,24 @@ learn_status_cb (void *opaque, const char *line)
 
   if (keywordlen == 8 && !memcmp (keyword, "SERIALNO", keywordlen))
     {
-      if (parm->serialno)
-	free (parm->serialno);
+      free (parm->serialno);
       parm->serialno = store_serialno (line);
     }
+  else if (keywordlen == 13 && !memcmp (keyword, "$DISPSERIALNO", keywordlen))
+    {
+      free (parm->dispserialno);
+      parm->dispserialno = unescape_status_string (line);
+    }
+  else if (keywordlen == 7 && !memcmp (keyword, "APPTYPE", keywordlen))
+    {
+      parm->is_piv = !strcmp (line, "piv");
+      parm->is_opgp = !strcmp (line, "openpgp");
+    }
+  else if (keywordlen == 8 && !memcmp (keyword, "CARDTYPE", keywordlen))
+    {
+      free (parm->cardtype);
+      parm->cardtype = unescape_status_string (line);
+    }
   else if (keywordlen == 9 && !memcmp (keyword, "DISP-NAME", keywordlen))
     {
       if (parm->disp_name)
@@ -572,38 +632,67 @@ learn_status_cb (void *opaque, const char *line)
     }
   else if (keywordlen == 11 && !memcmp (keyword, "KEYPAIRINFO", keywordlen))
     {
-      const char *grip = line;
+      /* The format of such a line is:
+       *   KEYPAIRINFO <hexgrip> <keyref> [<usage>]
+       */
+      const char *hexgrip = line;
+      char *line_buffer, *p;
+      const char *usage;
+
       while (*line && !spacep (line))
         line++;
+      while (spacep (line))
+        line++;
 
-      if (line - grip == 40)
-	{
-	  while (spacep (line))
-	    line++;
-	  if (!memcmp (line, "OPENPGP.", 8))
-	    {
-	      int no;
-	      line += 8;
+      p = line_buffer = strdup (line);
+      if (!line_buffer)
+        goto no_core;
+      keyref = line_buffer;
+      while (*p && !spacep (p))
+        p++;
+      if (*p)
+        {
+          *p++ = 0;
+          while (spacep (p))
+            p++;
+          usage = p;
+          while (*p && !spacep (p))
+            p++;
+          *p = 0;
+        }
+      else
+        usage = "";
 
-	      no = atoi (line);
+      if (hexgrip_valid_p (hexgrip))
+        {
+          /* Check whether we already have an item for the keyref.  */
+          kinfo = scute_find_kinfo (parm, keyref);
+          if (!kinfo)  /* New entry.  */
+            {
+              kinfo = create_kinfo (parm, keyref);
+              if (!kinfo)
+                goto no_core;
+            }
+          else /* Existing entry - clear grip and usage.  */
+            {
+              *kinfo->grip = 0;
+              memset (&kinfo->usage, 0, sizeof kinfo->usage);
+            }
 
-	      if (no == 1)
-		{
-		  memcpy (parm->grip1, grip, 40);
-		  parm->grip1valid = 1;
-		}
-	      else if (no == 2)
-		{
-		  memcpy (parm->grip2, grip, 40);
-		  parm->grip2valid = 1;
-		}
-	      else if (no == 3)
-		{
-		  memcpy (parm->grip3, grip, 40);
-		  parm->grip3valid = 1;
-		}
-	    }
-	}
+          strncpy (kinfo->grip, hexgrip, sizeof kinfo->grip);
+          kinfo->grip[sizeof kinfo->grip -1] = 0;
+          for (; *usage; usage++)
+            {
+              switch (*usage)
+                {
+                case 's': kinfo->usage.sign = 1; break;
+                case 'c': kinfo->usage.cert = 1; break;
+                case 'a': kinfo->usage.auth = 1; break;
+                case 'e': kinfo->usage.encr = 1; break;
+                }
+            }
+        }
+      free (line_buffer);
     }
   else if (keywordlen == 6 && !memcmp (keyword, "EXTCAP", keywordlen))
     {
@@ -630,6 +719,9 @@ learn_status_cb (void *opaque, const char *line)
         }
     }
   return 0;
+
+ no_core:
+  return gpg_error_from_syserror ();
 }
 
 
@@ -638,13 +730,62 @@ gpg_error_t
 scute_agent_learn (struct agent_card_info_s *info)
 {
   gpg_error_t err;
+  int has_opt_all = 0;
+
 
   memset (info, 0, sizeof (*info));
-  err = assuan_transact (agent_ctx, "LEARN --send",
-			 NULL, NULL, default_inq_cb,
-			 NULL, learn_status_cb, info);
+  err = ensure_agent_connection ();
+  if (!err && !agent_simple_cmd (agent_ctx,
+                                 "SCD GETINFO cmd_has_option SERIALNO all"))
+    has_opt_all = 1; /* SERIALNO --all and LEARN --multi is okay.  */
 
-  return err;
+  if (!err)
+    {
+      /* First do a serialno to reset the card-removed-flag and also
+       * to make sure that additional applications are enabled.  We do
+       * not check the error here as we catch that after the LEARN.  */
+      agent_simple_cmd (agent_ctx, (has_opt_all? "SCD SERIALNO --all"
+                                    /*       */: "SCD SERIALNO"      ));
+      err = assuan_transact (agent_ctx,
+                             (has_opt_all? "SCD LEARN --force --multi"
+                              /*       */: "SCD LEARN --force"),
+                             NULL, NULL,
+                             default_inq_cb, NULL,
+                             learn_status_cb, info);
+    }
+
+  if (gpg_err_source(err) == GPG_ERR_SOURCE_SCD
+      && gpg_err_code (err) == GPG_ERR_CARD_REMOVED)
+    {
+      /* SCD session is in card removed state.  clear that state.
+       * That should have been cleared by the initial SERIALNO but
+       * other processes may race with that.  */
+      err = assuan_transact (agent_ctx, "SCD SERIALNO",
+                             NULL, NULL, NULL, NULL, NULL, NULL);
+      if (!err)
+        {
+          memset (info, 0, sizeof (*info));
+          err = assuan_transact (agent_ctx,
+                                 (has_opt_all? "SCD LEARN --force --multi"
+                                  /*       */: "SCD LEARN --force"),
+                                 NULL, NULL,
+                                 default_inq_cb, NULL,
+                                 learn_status_cb, info);
+        }
+    }
+  if (!err)
+    {
+      /* Also try to get the human readable serial number.  */
+      err = assuan_transact (agent_ctx, "SCD GETATTR $DISPSERIALNO",
+                             NULL, NULL,
+                             default_inq_cb, NULL,
+                             learn_status_cb, info);
+      if (gpg_err_code (err) == GPG_ERR_INV_NAME
+          || gpg_err_code (err) == GPG_ERR_UNSUPPORTED_OPERATION)
+        err = 0; /* Not implemented or GETATTR not supported.  */
+    }
+
+  return check_broken_pipe (err);
 }
 
 
@@ -705,14 +846,21 @@ scute_agent_check_status (void)
   int any = 0;
   char flag = '-';
 
+  err = ensure_agent_connection ();
+  if (err)
+    return err;
+
   /* First we look at the eventcounter to see if anything happened at
      all.  This is a low overhead function which won't even clutter a
      gpg-agent log file.  There is no need for error checking here. */
   if (last_flag)
-    assuan_transact (agent_ctx, "GETEVENTCOUNTER",
-                     NULL, NULL,
-                     NULL, NULL,
-                     geteventcounter_status_cb, &any);
+    {
+      err = assuan_transact (agent_ctx, "GETEVENTCOUNTER",
+                             NULL, NULL,
+                             NULL, NULL,
+                             geteventcounter_status_cb, &any);
+      check_broken_pipe (err);
+    }
 
   if (any || !last_flag)
     {
@@ -720,6 +868,7 @@ scute_agent_check_status (void)
                              read_status_cb, &flag,
                              default_inq_cb, NULL,
                              NULL, NULL);
+      err = check_broken_pipe (err);
       if (err)
         return err;
       last_flag = flag;
@@ -765,7 +914,7 @@ pksign_cb (void *opaque, const void *buffer, size_t length)
 }
 
 /* Parse the result of an pksign operation which is a s-expression in
-   normal form that looks like (7:sig-val(3:rsa(1:s<LENGTH>:<DATA>))).
+   canonical form that looks like (7:sig-val(3:rsa(1:s<LENGTH>:<DATA>))).
    The raw result is stored in RESULT of size *LEN, and *LEN is
    adjusted to the actual size.  */
 static gpg_error_t
@@ -808,7 +957,7 @@ pksign_parse_result (const struct signature *sig,
   if (! n)
     return gpg_error (GPG_ERR_INV_SEXP);
 
-  /* Remove nul byte prepended by gpg-agent. */
+  /* Remove a possible prepended zero byte. */
   if (!*s && n > 1)
     {
       n -= 1;
@@ -901,9 +1050,13 @@ decode_hash (const unsigned char *data, int len,
   return 0;
 }
 
-/* Call the agent to learn about a smartcard.  */
+
+/* Call the agent to sign (DATA,LEN) using the key described by
+ * HEXGRIP.  Stores the signature in SIG_RESULT and its length at
+ * SIG_LEN; SIGLEN must initially point to the allocated size of
+ * SIG_RESULT.  */
 gpg_error_t
-scute_agent_sign (char *grip, unsigned char *data, int len,
+scute_agent_sign (const char *hexgrip, unsigned char *data, int len,
 		  unsigned char *sig_result, unsigned int *sig_len)
 {
   char cmd[150];
@@ -931,12 +1084,17 @@ scute_agent_sign (char *grip, unsigned char *data, int len,
       return 0;
     }
 
-  if (grip == NULL || sig_result == NULL)
+  if (!hexgrip || !sig_result)
     return gpg_error (GPG_ERR_INV_ARG);
 
-  snprintf (cmd, sizeof (cmd), "SIGKEY %s", grip);
+  snprintf (cmd, sizeof (cmd), "SIGKEY %s", hexgrip);
+
+  err = ensure_agent_connection ();
+  if (err)
+    return err;
   err = assuan_transact (agent_ctx, cmd, NULL, NULL, default_inq_cb,
-			 NULL, NULL, NULL);
+                         NULL, NULL, NULL);
+  err = check_broken_pipe (err);
   if (err)
     return err;
 
@@ -947,11 +1105,13 @@ scute_agent_sign (char *grip, unsigned char *data, int len,
   snprintf (cmd, sizeof (cmd), "SETHASH --hash=%s %s", hash, pretty_data);
   err = assuan_transact (agent_ctx, cmd, NULL, NULL, default_inq_cb,
 			 NULL, NULL, NULL);
+  err = check_broken_pipe (err);
   if (err)
     return err;
 
   err = assuan_transact (agent_ctx, "PKSIGN",
 			 pksign_cb, &sig, default_inq_cb, NULL, NULL, NULL);
+  err = check_broken_pipe (err);
   if (err)
     return err;
 
@@ -960,17 +1120,209 @@ scute_agent_sign (char *grip, unsigned char *data, int len,
 }
 
 
+
+struct pkdecrypt_parm_s
+{
+  unsigned int len;
+  unsigned char data[512];
+  assuan_context_t ctx;
+  const unsigned char *ciphertext;
+  size_t ciphertextlen;
+};
+
+
+static gpg_error_t
+pkdecrypt_data_cb (void *opaque, const void *buffer, size_t length)
+{
+  struct pkdecrypt_parm_s *parm = opaque;
+
+  if (parm->len + length > sizeof parm->data)
+    {
+      DEBUG (DBG_INFO, "maximum decryption result length exceeded");
+      return gpg_error (GPG_ERR_BAD_DATA);
+    }
+
+  memcpy (parm->data + parm->len, buffer, length);
+  parm->len += length;
+
+  return 0;
+}
+
+
+/* Handle the inquiries from pkdecrypt.  Note, we only send the data,
+ * assuan_transact takes care of flushing and writing the "END".  */
+static gpg_error_t
+pkdecrypt_inq_cb (void *opaque, const char *line)
+{
+  struct pkdecrypt_parm_s *parm = opaque;
+  gpg_error_t err;
+  const char *keyword = line;
+  int keywordlen;
+
+  for (keywordlen = 0; *line && !spacep (line); line++, keywordlen++)
+    ;
+  while (spacep (line))
+    line++;
+
+  if (keywordlen == 10 && !memcmp (keyword, "CIPHERTEXT", 10))
+    err = assuan_send_data (parm->ctx, parm->ciphertext, parm->ciphertextlen);
+  else
+    err = default_inq_cb (NULL, line);
+
+  return err;
+}
+
+
+
+/* Parse the result of a pkdecrypt operation which is an s-expression
+ * in canonical form that looks like
+ * (5:value<NDATA>:<DATA>).
+ *
+ * The raw result is stored in RESULT which has a size of *R_LEN, and
+ * *R_LEN is adjusted to the actual size.  */
+static gpg_error_t
+pkdecrypt_parse_result (struct pkdecrypt_parm_s *ctx,
+                        unsigned char *result, unsigned int *r_len)
+{
+  char *buf = ctx->data;
+  size_t len = ctx->len;
+  char *endp, *raw;
+  size_t n, rawlen;
+
+  if (len < 13 || memcmp (buf, "(5:value", 8) )
+    return gpg_error (GPG_ERR_INV_SEXP);
+  len -= 8;
+  buf += 8;
+
+  n = strtoul (buf, &endp, 10);
+  if (!n || *endp != ':')
+    return gpg_error (GPG_ERR_INV_SEXP);
+  endp++;
+  if ((endp-buf)+n > len)
+    return gpg_error (GPG_ERR_INV_SEXP); /* Oops: Inconsistent S-Exp. */
+
+  /* Let (RAW,RAWLEN) describe the pkcs#1 block and remove that padding.  */
+  raw = endp;
+  rawlen = n;
+
+  if (rawlen < 10)  /* 0x00 + 0x02 + <1_random> + 0x00 + <16-session> */
+    return gpg_error (GPG_ERR_INV_SESSION_KEY);
+
+  if (raw[0] || raw[1] != 2 )  /* Wrong block type version. */
+    return gpg_error (GPG_ERR_INV_SESSION_KEY);
+
+  for (n=2; n < rawlen && raw[n]; n++) /* Skip the random bytes. */
+    ;
+  if (n+1 >= rawlen || raw[n] )
+    return gpg_error (GPG_ERR_INV_SESSION_KEY);
+  n++; /* Skip the zero byte */
+
+  if (*r_len < (rawlen - n))
+    return gpg_error (GPG_ERR_TOO_LARGE);
+  memcpy (result, raw + n, rawlen - n);
+  *r_len = rawlen - n;
+  return 0;
+}
+
+
+/* Call the agent to decrypt (ENCDATA,ENCDATALEN) using the key
+ * described by HEXGRIP.  Stores the plaintext at R_PLAINDATA and its
+ * length at R_PLAINDATALEN; R_PLAINDATALEN must initially point to
+ * the allocated size of R_PLAINDATA and is updated to the actual used
+ * size on return.  */
+gpg_error_t
+scute_agent_decrypt (const char *hexgrip,
+                     unsigned char *encdata, int encdatalen,
+                     unsigned char *r_plaindata, unsigned int *r_plaindatalen)
+{
+  char cmd[150];
+  gpg_error_t err;
+  struct pkdecrypt_parm_s pkdecrypt;
+  char *s_data;
+  size_t s_datalen;
+
+  if (!hexgrip || !encdata || !encdatalen || !r_plaindatalen)
+    return gpg_error (GPG_ERR_INV_ARG);
+
+  if (!r_plaindata)
+    {
+      /* Fixme: We do not return the minimal required length but our
+       * internal buffer size.  */
+      pkdecrypt.len = *r_plaindatalen;
+      *r_plaindatalen = sizeof pkdecrypt.data - 1;
+      if (pkdecrypt.len > sizeof pkdecrypt.data - 1)
+        return gpg_error (GPG_ERR_INV_LENGTH);
+      return 0;
+    }
+
+  err = ensure_agent_connection ();
+  if (err)
+    return err;
+
+  snprintf (cmd, sizeof (cmd), "SETKEY %s", hexgrip);
+  err = assuan_transact (agent_ctx, cmd,
+                         NULL, NULL,
+                         default_inq_cb, NULL,
+			 NULL, NULL);
+  err = check_broken_pipe (err);
+  if (err)
+    return err;
+
+  /* Convert the input into an appropriate s-expression as expected by
+   * gpg-agent which is:
+   *
+   *  (enc-val
+   *    (flags pkcs1)
+   *    (rsa
+   *      (a VALUE)))
+   *
+   * Out of convenience we append a non-counted extra nul to the
+   * created canonical s-expression.
+   */
+  s_data = malloc (100 + encdatalen);
+  if (!s_data)
+    return gpg_error_from_syserror ();
+  snprintf (s_data, 50, "(7:enc-val(5:flags5:pkcs1)(3:rsa(1:a%d:",
+            encdatalen);
+  s_datalen = strlen (s_data);
+  memcpy (s_data + s_datalen, encdata, encdatalen);
+  s_datalen += encdatalen;
+  memcpy (s_data + s_datalen, ")))", 4);
+  s_datalen += 3;
+
+  pkdecrypt.len = 0;
+  pkdecrypt.ctx = agent_ctx;
+  pkdecrypt.ciphertext = s_data;
+  pkdecrypt.ciphertextlen = s_datalen;
+  err = assuan_transact (agent_ctx, "PKDECRYPT",
+			 pkdecrypt_data_cb, &pkdecrypt,
+                         pkdecrypt_inq_cb, &pkdecrypt,
+                         NULL, NULL);
+  err = check_broken_pipe (err);
+  if (!err)
+    err = pkdecrypt_parse_result (&pkdecrypt, r_plaindata, r_plaindatalen);
+
+  free (s_data);
+  return err;
+}
+
+
 /* Determine if FPR is trusted.  */
 gpg_error_t
-scute_agent_is_trusted (char *fpr, bool *is_trusted)
+scute_agent_is_trusted (const char *fpr, bool *is_trusted)
 {
   gpg_error_t err;
   bool trusted = false;
   char cmd[150];
 
+  err = ensure_agent_connection ();
+  if (err)
+    return err;
+
   snprintf (cmd, sizeof (cmd), "ISTRUSTED %s", fpr);
   err = assuan_transact (agent_ctx, cmd, NULL, NULL, default_inq_cb,
 			 NULL, NULL, NULL);
+  err = check_broken_pipe (err);
   if (err && gpg_err_code (err) != GPG_ERR_NOT_TRUSTED)
     return err;
   else if (!err)
@@ -1027,9 +1379,9 @@ get_cert_data_cb (void *opaque, const void *data, size_t data_len)
 }
 
 
-/* Try to get certificate for key numer NO.  */
+/* Try to get certificate for CERTREF.  */
 gpg_error_t
-scute_agent_get_cert (int no, struct cert *cert)
+scute_agent_get_cert (const char *certref, struct cert *cert)
 {
   gpg_error_t err;
   char cmd[150];
@@ -1039,9 +1391,14 @@ scute_agent_get_cert (int no, struct cert *cert)
   cert_s.cert_der_len = 0;
   cert_s.cert_der_size = 0;
 
-  snprintf (cmd, sizeof (cmd), "SCD READCERT OPENPGP.%i", no);
+  err = ensure_agent_connection ();
+  if (err)
+    return err;
+
+  snprintf (cmd, sizeof (cmd), "SCD READCERT %s", certref);
   err = assuan_transact (agent_ctx, cmd, get_cert_data_cb, &cert_s,
 			 NULL, NULL, NULL, NULL);
+  err = check_broken_pipe (err);
   /* Just to be safe... */
   if (!err && (cert_s.cert_der_len <= 16 || cert_s.cert_der[0] != 0x30))
     {
@@ -1060,6 +1417,8 @@ scute_agent_get_cert (int no, struct cert *cert)
 
   cert->cert_der = cert_s.cert_der;
   cert->cert_der_len = cert_s.cert_der_len;
+  strncpy (cert->certref, certref, sizeof cert->certref -1);
+  cert->certref[sizeof cert->certref - 1] = 0;
 
   return 0;
 }
@@ -1090,13 +1449,17 @@ scute_agent_get_random (unsigned char *data, size_t len)
     gpg_error_t err;
     struct random_request request;
 
+    err = ensure_agent_connection ();
+    if (err)
+      return err;
+
     snprintf (command, sizeof(command), "SCD RANDOM %zu", len);
 
     request.buffer = data;
     request.len = len;
     err = assuan_transact (agent_ctx, command, get_challenge_data_cb,
                            &request, NULL, NULL, NULL, NULL);
-
+    err = check_broken_pipe (err);
     return err;
 }
 
