@@ -1,31 +1,22 @@
 /* p11-findobjectsinit.c - Cryptoki implementation.
-   Copyright (C) 2006 g10 Code GmbH
-
-   This file is part of Scute.
- 
-   Scute is free software; you can redistribute it and/or modify it
-   under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
-   (at your option) any later version.
-
-   Scute is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with Scute; if not, write to the Free Software Foundation,
-   Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
-
-   In addition, as a special exception, g10 Code GmbH gives permission
-   to link this library: with the Mozilla Foundation's code for
-   Mozilla (or with modified versions of it that use the same license
-   as the "Mozilla" code), and distribute the linked executables.  You
-   must obey the GNU General Public License in all respects for all of
-   the code used other than "Mozilla".  If you modify this file, you
-   may extend this exception to your version of the file, but you are
-   not obligated to do so.  If you do not wish to do so, delete this
-   exception statement from your version.  */
+ * Copyright (C) 2006 g10 Code GmbH
+ *
+ * This file is part of Scute.
+ *
+ * Scute is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as
+ * published by the Free Software Foundation; either version 2.1 of
+ * the License, or (at your option) any later version.
+ *
+ * Scute is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this program; if not, see <https://gnu.org/licenses/>.
+ * SPDX-License-Identifier: LGPL-2.1-or-later
+ */
 
 #if HAVE_CONFIG_H
 #include <config.h>
@@ -41,9 +32,10 @@
 #include "error-mapping.h"
 #include "slots.h"
 
-
-CK_DEFINE_FUNCTION(CK_RV, C_FindObjectsInit)
-     (CK_SESSION_HANDLE hSession, CK_ATTRIBUTE_PTR pTemplate, CK_ULONG ulCount)
+
+CK_RV CK_SPEC
+C_FindObjectsInit (CK_SESSION_HANDLE hSession,
+                   CK_ATTRIBUTE_PTR pTemplate, CK_ULONG ulCount)
 {
   CK_RV err = CKR_OK;
   slot_iterator_t slot;
@@ -128,6 +120,18 @@ CK_DEFINE_FUNCTION(CK_RV, C_FindObjectsInit)
       goto out;
     }
 
+  if (!search_result_len)
+    {
+      /* We do not yet known about this object.  If CKA_ISSUER and
+       * CKA_SERIAL_NUMBER was requested, try to look it up via
+       * gpgsm.  The way we can implement this would be a new option to
+       * gpgsm's LISTKEYS, named "--der" which takes the raw DER
+       * encoding of both items.  The advantage of doing this in gpgsm
+       * is that we can use libksba there to build the actual
+       * rfc-2253 search string from the DER. */
+
+    }
+
   err = session_set_search_result (slot, session, search_result,
 				   search_result_len);
 
